"""
relvar.py – TclRAL operations on relvars
"""

import logging
from typing import List, Dict, Any, Optional
from pyral.database import Database
from pyral.transaction import Transaction
from pyral.relation import Relation, _relation
from pyral.rtypes import Attribute, Mult, delim, RelationValue, body, header, snake
from collections import namedtuple

_logger = logging.getLogger(__name__)


class Relvar:
    """
    A relational variable (table)

    TclRAL does not support spaces in names, but PyRAL accepts space delimited names.
    But each space delimiter will be replaced with an underscore delimiter before submitting to TclRAL
    """

    # TclRAL commands organized in alphabetic order to match the TclRAL man pages
    @classmethod
    def printall(cls, db: str):
        """
        Print out all relvar relations in the tclRAL database in alphabetic order

        :param db: DB session name
        """
        cmd = 'relvar names'  # TclRAL returns all relvar names as a single string
        # strip off first '::' and then split the rest on ' ::'
        relvar_names = Database.execute(db, cmd).lstrip('::').split(' ::')
        if relvar_names:
            relvar_names.sort()
            for r in relvar_names:
                Relation.print(db, r)
        else:
            print("No relvars to print")

    @classmethod
    def create_association(cls, db: str, name: str,
                           from_relvar: str, from_attrs: List[str], from_mult: Mult,
                           to_relvar: str, to_attrs: List[str], to_mult: Mult,
                           ):
        """
        Create a named referential association between two relvars. TclRAL returns an empty string and PyRAL
        does not return any value.

        An association declares that a source relvar has attributes that refer to an identifier of a target relvar.
        In this method we'll use the terms "from" and "to" to describe the corresponding association elements.

        The referring and referred to attribute lists are ordered and "from" attributes refer to the corresponding
        "to" attributes.

        The set of "to" attributes must constitute an identifier for the "to" relvar.

        Multiplicity and conditionality are specified as M, 1, Mc or 1c with the meanings:

            M - at least one
            1 - exactly one
            Mc - zero or more
            1c - zero or one

        :param db: DB session name
        :param name: Name of the association, an association Rnum in SM xUML
        :param from_relvar: The referring relvar (source)
        :param from_attrs: The referential attributes in the source relvar
        :param from_mult: Multiplicity conditionality associated with the source side of the association
        :param to_relvar: The referenced relvar (target)
        :param to_attrs: The referential attributes in the target relvar
        :param to_mult: Multiplicity conditionality associated with the target side of the association
        """
        # Join each attribute list into a string
        from_attr_str = "{" + ' '.join(from_attrs) + '}'
        to_attr_str = "{" + ' '.join(to_attrs) + '}'

        # Build a TclRAL command string
        cmd = f"relvar association {name} {from_relvar} {from_attr_str} {from_mult.value}" \
              f" {to_relvar} {to_attr_str} {to_mult.value}"

        # Execute the command and ignore the result
        Database.execute(db, cmd=cmd, log=False)
        # Verify and log the constraint by executing the TclRAL constraint command
        verify_cmd = f"relvar constraint info {name}"
        Database.execute(db, cmd=verify_cmd)

    @classmethod
    def create_correlation(cls, db: str, name: str, correlation_relvar: str,
                           correl_a_attrs: List[str], a_mult: Mult, a_relvar: str, a_ref_attrs: List[str],
                           correl_b_attrs: List[str], b_mult: Mult, b_relvar: str, b_ref_attrs: List[str],
                           complete: bool = False):
        """
        A relvar can manage a correlation between tuples in one or two other relvars.

        An association class in Shlaer-Mellor xUML is represented by such a relvar correlation.

        The referential correlation is between a correlation relvar and two other relvars A and B.

        Correlations declare that every tuple in the correlation relvar refers to exactly one tuple of
        relvar A and exactly one tuple of relvar B. (we use lower case a and b here in the code)

        The correlation A attributes refer to an identifier of relvar A and similarly for the B components.

        Multiplicity and conditionality is specified the same as in the create_association command.

        Syntax from the TclRAL man page:
        relvar correlation ?-complete? name correlRelvar
             correlAttrListA refToSpecA refToRelvarA refToAttrListA
             correlAttrListB refToSpecB refToRelvarB refToAttrListB

        Example TclRAL command from the TclRAL man page
        PyRAL:

            relvar correlation C1 OWNERSHIP
                OwnerName + OWNER OwnerName
                DogName * DOG DogName

        Additional example from the SM metamodel
            (See SM class-attribute subsystem with participating classesIdentifier and Attribute where
            the R22 association is:
                Identifier requires M Attribute / Attribute is required in Mc Identifier
                using association class Identifier Attribute):

            PyRAL input:
                name: R22, correlation_relvar: Identifier_Attribute
                corel_a_attrs: ['Identifier', 'Class', 'Domain']
                a_mult: Mc
                a_ref_attrs: ['Number', 'Class', 'Domain']
                corel_b_attrs: ['Attribute', 'Class', 'Domain']
                b_mult: M
                b_ref_attrs: ['Name', 'Class', 'Domain']

            TclRAL generated command
                relvar correlation R22 Identifier_Attribute
                    {Identifier Class Domain} + Identifier
                    {Number Class Domain} {Attribute Class Domain} * Attribute {Name Class Domain}

        :param db: DB session name
        :param name: Name of the correlation
        :param correlation_relvar: Name of the relvar holding the correlation
        :param correl_a_attrs: Attrs in correlation relvar referencing a-side relvar
        :param a_mult: Multiplicity on the a-side relvar
        :param a_relvar: Name of the a-side relvar
        :param a_ref_attrs: Attrs in the a-side relvar referencd by the correlation
        :param correl_b_attrs: Attrs in correlation relvar referencing b-side relvar
        :param b_mult: Multiplicity on the b-side relvar
        :param b_relvar: Name of the b-side relvar
        :param b_ref_attrs: Attrs in the b-side relvar referencd by the correlation
        :param complete: True implies the cardinality of correlRelvar must equal the product of the cardinality of
            refToRelvarA and refToRelvarB. If False, correlRelvar is allowed to have a subset of the Cartesian product
            of the references.
        :return:
        """
        # Join each attribute list into a string
        correl_a_attrs_str = "{" + ' '.join(correl_a_attrs) + '}'
        correl_b_attrs_str = "{" + ' '.join(correl_b_attrs) + '}'
        a_ref_attrs_str = "{" + ' '.join(a_ref_attrs) + '}'
        b_ref_attrs_str = "{" + ' '.join(b_ref_attrs) + '}'

        # Build a TclRAL command string
        # We need to reverse the a/b multiplicities since TclRAL considers multiplicty from the perspective
        # of the correlation relvar tuples rather than from the perspectives of the participating (non correlation)
        # relvar tuples. The latter approach matches the way Shlaer-Mellor xUML modelers think.
        cmd = f"relvar correlation {'-complete ' if complete else ''}{name} {correlation_relvar} " \
              f"{correl_a_attrs_str} {b_mult.value} {a_relvar} {a_ref_attrs_str} " \
              f"{correl_b_attrs_str} {a_mult.value} {b_relvar} {b_ref_attrs_str}"

        # Execute the command and log the result
        Database.execute(db, cmd=cmd, log=False)
        # Verify and log the constraint by executing the TclRAL constraint command
        verify_cmd = f"relvar constraint info {name}"
        Database.execute(db, cmd=verify_cmd)

    @classmethod
    def set(cls, db: str, relvar: str, relation: str = _relation, svar_name: Optional[str] = None) -> RelationValue:
        """

        From the TclRAL documentation:

        The set subcommand replaces the current value held by the relation variable named relvarName with the value
        given by relationValue.

        It is an error to attempt to assign a relation value to a relation variable that is of a different type
        than the type of the value that the variable currently holds. In other words, it is not possible to change
        the type of a relvar by assignment.

        The return value of the subcommand is the current value held by relvarName.

        If the relationValue argument is missing, then no attempt is made to change the value of relvarName.

        :param db: DB session name
        :param relvar: The name of an existing relvar
        :param relation: This relation becomes the new value of the relvar
        :param svar_name: An optional session variable that holds the result
        """
        cmd = f"set {_relation} [relvar set {{{relvar}}} ${{{relation}}}]"

        result = Database.execute(db=db, cmd=cmd)
        if svar_name:
            cls.set_var(db=db, name=svar_name)
        return Relation.make_pyrel(result)

    @classmethod
    def insert(cls, db: str, relvar: str, tuples: List[namedtuple], tr: Optional[str] = None):
        """
        Insert a set of tuples into the value of a relvar, modifying it in place.

        The tuples are concatenated into a single command that is either executed immediately
        or added to an open transaction if tr_name is provided.

        The TclRAL syntax is:
            relvar insert <relvarName> ?<name-value-list> ...?

        TclRAL example where an instance of Class is inserted into the SM xUML metamodel:
            relvar insert Class {Name {Accessible Shaft Level} Cnum {C1} Domain {Elevator Management}}

        Where PyRAL input is:
            relvar: 'Class'
            tuples: [ Class_i(Name='Accessible Shaft Level', Cnum='C1', Domain='Elevator Management') ]

        Note that the emtpy set may be provided in the TclRAL command which results in a no-op.
        PyRAL supports this feature by allowing an empty list of tuples to be specified

        :param db: DB session name
        :param relvar: The name of an existing relvar
        :param tuples: A list of tuples named such that the attributes exactly match the relvar header
        :param tr:  Optional transaction name, add to this transaction if supplied
        """
        b = body(tuples)

        # Start command with the relvar command prefix
        cmd = f"relvar insert {snake(relvar)} {b}"

        # Add to open transaction if tr_name is provided
        if tr:
            Transaction.append_statement(db=db, name=tr, statement=cmd)
        else:
            Database.execute(db=db, cmd=cmd)

    @classmethod
    def create_partition(cls, db: str, name: str,
                         superclass_name: str, super_attrs: List[str], subs: Dict[str, List[str]]):
        """
        A partition is defined such that the set of tuples in a super relvar is referenced by
        the tuples distributed across a set of one or more sub relvars. Each tuple in a sub
        relvar references some tuple in the super relvar and each tuple in the super relvar is
        referenced by one tuple in one of its sub relvars.

        This constraint implies that the tuples in the super relvar are completely partitioned
        into the disjoint sub sets given by the complete set of sub relvars.

        This constraint is used in SM xUML to support generalization relationships where
        a superclass instance population is paritioned across two or more subclass instance
        populations.

        The TclRAL syntax is:
            relvar partition <name> <super> <superAttrList>
                <sub1> <sub1AttrList>
                ...

        A TclRAL command example is:
            relvar partition R14 Subsystem_Element {Label Domain} Relationship {Rnum Domain} Class {Cnum Domain}

        This is generated from the PyRAL input:
            name: 'R14'
            superclass_name: 'Subsystem Element'
            super_attrs: ['Label', 'Domain']
            subs: {'Relationship': ['Rnum', 'Domain'], 'Class':['Cnum', 'Domain']}

        In the above examples both Relationship.Rnum and Class.Cnum refer to the Subsystem Element.Label
        attribute. So the ordering of identifier attributes for each relvar is significant.

        :param db: DB session name
        :param name: Name of the partition, a generalizaiton rnum in SM xUML
        :param superclass_name: Name of the superclass relvar
        :param super_attrs: A list of attributes constituting an identifier of the superclass referenced by
        :param subs:
        """
        super_attrs_str = '{' + ' '.join(super_attrs) + '}'
        all_subs = ""
        for subname, attrs in subs.items():
            all_subs += subname + ' ' + '{' + ' '.join(attrs) + '} '
        all_subs = all_subs[:-1]

        cmd = f"relvar partition {name} {superclass_name} {super_attrs_str} {all_subs}"
        # Execute the command and log the result
        Database.execute(db, cmd=cmd, log=False)
        # Verify and log the constraint by executing the TclRAL constraint command
        verify_cmd = f"relvar constraint info {name}"
        Database.execute(db, cmd=verify_cmd)

    @classmethod
    def create_relvar(cls, db: str, name: str, attrs: List[Attribute], ids: Dict[int, List[str]]) -> str:
        """
        Create a relvar

        Syntax from the TclRAL man page:
            relvar create <relvarName> <heading> <id1> ?id2 ...?

        Example TclRAL command:
            relvar create Waypoint {WPT_number int, Lat string, Long string, Frequency double} {WPT_number} {Lat Long}

        This class has both a single attribute identifier "WPT_number" and a multiple attribute identifier {Lat Long}
        We wrap each identifier in {} brackets for simplicity even though we only really need them to group
        multiple attribute identifiers

        :param db: DB session name
        :param name: Name of the new relvar
        :param attrs: A list of Attributes (name, type - named tuples)
        :param ids: A dictionary of {idnum: [attr_name, ...] } values
        :return: The relation defined by the empty relvar in the form: <heading> {}
        """
        h = header(attrs)

        # Now we make the list of identifiers such as:
        #   {WPT_number} {Lat Long}
        id_list = ""
        for inum, attrs in ids.items():
            # Create a bracketed list for each identifier
            id_list += '{'
            for a in attrs:
                id_list += f"{a.replace(' ', delim)} "
            id_list = id_list[:-1] + '} '
        id_list = id_list[:-1]

        # Build and execute the TclRAL command
        cmd = f"relvar create {snake(name)} {h} {id_list}"
        return Database.execute(db=db, cmd=cmd)

    @classmethod
    def updateone(cls, db: str, relvar_name: str, id: Dict, update: Dict[str, Any]):
        """
        Modifies in place at most one tuple of the relvar's value.

        TclRAL syntax:
            relvar updateone <relvarName> <tupleVarName> <id-name-value-list> <script>

        Here is an example where an instance of Attribute in the SM Metamodel has its Type attribute updated:
        The TclRAL command (all on one line, but idented for readability here) is:
            relvar updateone Attribute t
                {Name {Floor} Class {Accessible Shaft Level} Domain {Elevator Management} }
                {tuple update $t Type {Level Name}}

        Generated from the PyRAL:
            relvar_name: 'Attribute'
            id: {'Name': 'Floor', 'Class': 'Accessible Shaft Level', 'Domain': 'Elevator Management'}
            update: {'Type': 'Level Name'}

        :param db: DB session name
        :param relvar_name: The relvar to be updated
        :param id: Identifier value for the tuple to be updated
        :param update: A dictionary of attribute value pairs whose values will be applied
        :return: A relation value with the same heading as the value held in relvarName and whose body contains either
        the single tuple that was updated or is empty if no matching tuple was found.
        """
        id_str = ""
        for id_attr, id_val in id.items():
            id_str += f"{id_attr} {{{id_val}}} "
        update_str = ""
        for u_attr, u_val in update.items():
            update_str += u_attr + " {" + u_val + "}"
        cmd = f'relvar updateone {relvar_name} t {{{id_str}}} {{tuple update $t {update_str}}}'
        return Database.execute(db, cmd)

    @classmethod
    def deleteone(cls, db: str, relvar_name: str, tid: Dict, tr: Optional[str] = None) -> str:
        """
        Deletes in place at most one tuple of the relvar's value.

        TclRAL syntax:
            relvar deleteone <relvarName> <tupleVarName> <id-name-value-list> <script>

        Here is an example where an instance of Attribute in the SM Metamodel has its Type attribute updated:
        The TclRAL command (all on one line, but idented for readability here) is:
            relvar deleteone Attribute t
                {Name {Floor} Class {Accessible Shaft Level} Domain {Elevator Management} }
                {tuple update $t Type {Level Name}}

        Generated from the PyRAL:
            relvar_name: 'Attribute'
            tid: {'Name': 'Floor', 'Class': 'Accessible Shaft Level', 'Domain': 'Elevator Management'}

        :param db: DB session name
        :param relvar_name: The relvar to be deleted
        :param tid: Identifier value for the tuple to be deleted
        :param tr: If name provided, appended to transaction, otherwise execute now and return result
        :return: A relation value with the same heading as the value held in relvarName and whose body contains either
        the single tuple that was updated or is empty if no matching tuple was found.
        """
        id_str = ""
        for id_attr, id_val in tid.items():
            id_str += f"{id_attr} {{{id_val}}} "
        cmd = f'relvar deleteone {relvar_name} {id_str}'
        if not tr:
            return Database.execute(db, cmd)
        else:
            Transaction.append_statement(db, name=tr, statement=cmd)
            return ''

    @classmethod
    def select_id(cls, db: str, relvar_name: str, tid: Dict, svar_name: Optional[str] = None) -> RelationValue:
        """
        Deletes in place at most one tuple of the relvar's value.

        TclRAL syntax:
            relvar deleteone <relvarName> <tupleVarName> <id-name-value-list> <script>

        Here is an example where an instance of Attribute in the SM Metamodel has its Type attribute updated:
        The TclRAL command (all on one line, but idented for readability here) is:
            relvar deleteone Attribute t
                {Name {Floor} Class {Accessible Shaft Level} Domain {Elevator Management} }
                {tuple update $t Type {Level Name}}

        Generated from the PyRAL:
            relvar_name: 'Attribute'
            tid: {'Name': 'Floor', 'Class': 'Accessible Shaft Level', 'Domain': 'Elevator Management'}

        :param db: DB session name
        :param svar_name:
        :param relvar_name: The relvar to be deleted
        :param tid: Identifier value for the tuple to be deleted
        :return: A relation value with the same heading as the value held in relvarName and whose body contains either
        the single tuple that was updated or is empty if no matching tuple was found.
        """
        id_str = ""
        for id_attr, id_val in tid.items():
            id_str += f"{id_attr} {{{id_val}}} "
        cmd = f'set {_relation} [relvar restrictone {snake(relvar_name)} {id_str}]'
        result = Database.execute(db, cmd)
        if svar_name:  # Save the result using the supplied session variable name
            Relation.set_var(db, svar_name)
        return Relation.make_pyrel(result, name=svar_name)



"""
dogs_example.py -- Set up basic dogs example

"""
# System
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction
from pyral.rtypes import Attribute


DOG_i = namedtuple('DOG_i', 'DogName Breed')
OWNER_i = namedtuple('OWNER_i', 'OwnerName Age City')
OWNERSHIP_i = namedtuple('OWNERSHIP_i', 'OwnerName DogName Acquired')

ddb = "dogs"

class Dogs:
    @classmethod
    def setup(cls):
        Database.open_session(name=ddb)
        Relvar.create_relvar(db=ddb, name="DOG", attrs=[
            Attribute(name="DogName", type="string"),
            Attribute(name="Breed", type="string"),
        ], ids={1: ["DogName"]})
        Relvar.create_relvar(db=ddb, name="OWNER", attrs=[
            Attribute(name="OwnerName", type="string"),
            Attribute(name="Age", type="int"),
            Attribute(name="City", type="string"),
        ], ids={1: ["OwnerName"]})
        Relvar.create_relvar(db=ddb, name="OWNERSHIP", attrs=[
            Attribute(name="DogName", type="string"),
            Attribute(name="OwnerName", type="string"),
            Attribute(name="Acquired", type="string"),
        ], ids={1: ["OwnerName", "DogName"]})
        Relvar.insert(db=ddb, relvar='DOG', tuples=[
            DOG_i(DogName="Fido", Breed="Poodle"),
            DOG_i(DogName="Sam", Breed = "Collie"),
            DOG_i(DogName="Spot", Breed = "Terrier"),
            DOG_i(DogName="Rover", Breed = "Retriever"),
            DOG_i(DogName="Fred", Breed = "Spaniel"),
            DOG_i(DogName="Jumper", Breed = "Mutt"),
        ])
        Relvar.insert(db=ddb, relvar='OWNER', tuples=[
            OWNER_i(OwnerName="Sue", Age=24, City="Cupertino"),
            OWNER_i(OwnerName="George", Age=35, City="Sunnyvale"),
            OWNER_i(OwnerName="Alice", Age=30, City="Cupertino"),
            OWNER_i(OwnerName="Mike", Age=50, City="San Jose"),
            OWNER_i(OwnerName="Jim", Age=42, City="San Francisco"),
        ])
        Relvar.insert(db=ddb, relvar='OWNERSHIP', tuples=[
            OWNERSHIP_i(OwnerName="Sue", DogName="Fido", Acquired="2001"),
            OWNERSHIP_i(OwnerName="Sue", DogName="Sam", Acquired="2000"),
            OWNERSHIP_i(OwnerName="George", DogName="Fido", Acquired="2001"),
            OWNERSHIP_i(OwnerName="George", DogName="Sam", Acquired="2000"),
            OWNERSHIP_i(OwnerName="Alice", DogName="Spot", Acquired="2001"),
            OWNERSHIP_i(OwnerName="Mike", DogName="Rover", Acquired="2002"),
            OWNERSHIP_i(OwnerName="Jim", DogName="Fred", Acquired="2003"),
        ])

        Relvar.printall(db=ddb)



"""
sum_play.py -- Play around with summarization

"""
# System
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction
from pyral.rtypes import Attribute


Flow_Dependency_i = namedtuple('Flow_Dependency_i', 'From_action To_action')
Actionf_i = namedtuple('Actionf_i', 'From_action')
Actionta_i = namedtuple('Actionta_i', 'To_action')
Action_i = namedtuple('Action_i', 'ID')

fdb = "fdb"  # Flow database example

class SumTest:
    """
    Summarization example
    """

    @classmethod
    def setup(cls):
        """
        Initialize the example
        """
        Database.open_session(name=fdb)
        Relvar.create_relvar(db=fdb, name='Flow_Dependency', attrs=[
            Attribute('From_action', 'string'),
            Attribute('To_action', 'string'),
        ], ids={1: ['From_action', 'To_action']})

        Relvar.insert(db=fdb, relvar='Flow_Dependency', tuples=[
            Flow_Dependency_i(From_action="ACTN1", To_action="ACTN4"),
            Flow_Dependency_i(From_action="ACTN1", To_action="ACTN2"),
            Flow_Dependency_i(From_action="ACTN2", To_action="ACTN3"),
            Flow_Dependency_i(From_action="ACTN3", To_action="ACTN7"),
            Flow_Dependency_i(From_action="ACTN4", To_action="ACTN5"),
            Flow_Dependency_i(From_action="ACTN5", To_action="ACTN6"),
            Flow_Dependency_i(From_action="ACTN6", To_action="ACTN7"),
            Flow_Dependency_i(From_action="ACTN7", To_action="ACTN9"),
            Flow_Dependency_i(From_action="ACTN8", To_action="ACTN9"),
            Flow_Dependency_i(From_action="ACTN9", To_action="ACTN13"),
            Flow_Dependency_i(From_action="ACTN9", To_action="ACTN12"),
            Flow_Dependency_i(From_action="ACTN10", To_action="ACTN11"),
            Flow_Dependency_i(From_action="ACTN11", To_action="ACTN13"),
            Flow_Dependency_i(From_action="ACTN11", To_action="ACTN12"),
        ])

        Relvar.printall(fdb)

        cls.play()

    @classmethod
    def play(cls):

        # Specify a set of initial from actions that have completed execution
        Relation.create(db=fdb, attrs=[Attribute(name="From_action", type="string")],
                        tuples=[
                            Actionf_i(From_action="ACTN1"),
                            Actionf_i(From_action="ACTN10"),
                            Actionf_i(From_action="ACTN8"),
                        ], svar_name="xactions")
        Relation.print(db=fdb, variable_name="xactions")

        # Find all downstream dependencies
        Relation.join(db=fdb, rname1="Flow_Dependency", rname2="xactions")
        Relation.project(db=fdb, attributes=("To_action",), svar_name="downstream")
        Relation.print(db=fdb, variable_name="downstream")

        # Find all required upstream inputs for the downstream actions
        Relation.join(db=fdb, rname1="Flow_Dependency", rname2="downstream", svar_name="required_inputs")
        Relation.print(db=fdb, variable_name="required_inputs")

        # Add a boolean attribute named Can_execute
        # For each downstream action (per To_action) join it back to required inputs, projecting on the From_action
        # This yields the set of upstream executed actions for that one downstream action
        # Then test to see if this set is a subset of the xactions
        # If so, the downstream action has all dependencies fulfilled and can now execute (true)
        # Otherwise, there are some required input actions that have not just executed
        Relation.raw(
            db=fdb, cmd_str=r"relation summarizeby $required_inputs To_action s Can_execute boolean {\
            [relation is [relation project [relation join $required_inputs $s] From_action] subsetof $xactions]}",
            svar_name="c"
        )
        Relation.print(db=fdb, variable_name="c")





"""
rtypes.py -- Relational database types such as header, attribute, tuple, etc
"""
# System
from typing import List
from collections import namedtuple
from enum import Enum

Attribute = namedtuple('_Attribute', 'name type')
RelationValue = namedtuple('RelationValue', 'name header body')
SumExpr = namedtuple("SumExpr", "attr expr")
delim = '_' # TclRAL delimiter that replaces a space delimiter

# Commands
JoinCmd = namedtuple('JoinCmd', 'rname1 rname2 attrs')
SetCompareCmd = namedtuple('SetCompareCmd', 'rname1 rname2 op')
ProjectCmd = namedtuple('ProjectCmd', 'relation attributes')


Cardinality = "relation cardinality $s"


class SetOp(Enum):
    subset = 'subsetof'
    superset = 'supersetof'
    psubset = 'propersubsetof'
    psuperset = 'propersupersetof'
    eq = 'equal'
    equal = 'equal'
    neq = 'notequal'
    notequal = 'notequal'


class Extent(Enum):
    GREATEST = "greatest"
    LEAST = "least"

class Card(Enum):
    ONE = "one"
    ALL = "all"

class Order(Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"

class Mult(Enum):
    AT_LEAST_ONE = '+'
    EXACTLY_ONE = '1'
    ZERO_ONE_OR_MANY = '*'
    ZERO_OR_ONE = '?'


def snake(name: str) -> str:
    return name.replace(' ', '_')

def header(attrs: List[Attribute]) -> str:
    """
    Creates a header as a a bracketed list of attribute name pairs
    that can be spedified as part of a TclRAL command.

    Here's an example:
    {WPT_number int, Lat string, Long string, Frequency double}

    :param attrs: A tuple of attributes (name, type) pairs
    :return: A TclRAL header string
    """
    header_text = "{"
    for a in attrs:
        # We need to replace any spaces in an attribute name with underscores
        header_text += f"{a.name.replace(' ', delim)} {a.type.replace(' ', delim)} "
    return header_text[:-1] + "}"  # Replace the trailing space with a closing bracket

def body(tuples: List[namedtuple]) -> str:
    """
    A body is a set of tuples formated into a single line TclRAL string.

    Here's an example:

    {WPT_number 3 Lat {37° 46' 30" N} Long {-122° 25' 10"} Frequency 117.95}

    :return: A single line of text with a string of TclRAL tuples
    """
    # Add in all of the tuples
    body_text = ""
    for t in tuples:
        body_text += '{'
        instance_tuple = t._asdict()
        for k, v in instance_tuple.items():
            body_text += f"{k} {{{v}}} "  # Spaces are allowed in values
        body_text = body_text[:-1] + '} '
    return body_text[:-1]



"""
relation.py – Operations on relations
"""

# System
import logging
import re
from tabulate import tabulate
from typing import List, Optional, Dict, Tuple
from collections import namedtuple

# PyRAL
from pyral.rtypes import RelationValue, Attribute, header, body, SetOp, SumExpr, snake, Order, Card, Extent
from pyral.database import Database

_logger = logging.getLogger(__name__)

# If we want to apply successive (nested) operations in TclRAL we need to have the result
# of each TclRAL command saved in tcl variable. So each time we execute a command that produces
# a relation result we save it. The variable name is chosen so that it shouldn't conflict with
# any user relvars. Do not ever use the name below as one of your user relvars!
# For any given command, if no relvar is specified, the previous relation result is assumed
# to be the input.
_relation = r'^relation'  # Name of the latest relation result. Carat prevents name collision
_RANK = "_rank"  # Default name of the rank attribute added by extension using the rank command
_TAG = "_tag" # Default name of the tag attribute added by the tag command
session_variable_names = set()  # Maintain a list of temporary variable names in use


class Relation:
    """
    A relational value
    """

    @classmethod
    def declare_rv(cls, db: str, owner: str, name: str) -> str:
        """
        Saves

        :param db:
        :param owner:
        :param name:
        :return:
        """
        # Verify that db session exists
        if db not in Database.sessions:
            raise KeyError(f"Database session '{db}' has not been initialized.")

        db_rvs = Database.rv_names.setdefault(db,{})
        owner_rvs = db_rvs.setdefault(owner, set())

        if name in owner_rvs:
            raise KeyError(f"Relational variable {name} already defined for owner {owner}")

        owner_rvs.add(name)
        return f"{owner}__{name}"

    @classmethod
    def free_rvs(cls, db: str, owner: str, names: tuple[str, ...] = (), exclude: bool = False):
        """
        Unset relation variable names declared by the owner.

        :param db: Database session name
        :param owner: Name of the owner who declared the relational variables
        :param names: Names to include or exclude (depending on `exclude`)
        :param exclude: If True, *keep* the listed names and delete all others.
                        If False, *delete only* the listed names. If names is empty, delete all.
        """
        try:
            owner_rvs = Database.rv_names[db][owner]
        except KeyError:
            raise KeyError(f"No such owner '{owner}' in session '{db}'")

        names_to_remove = (
            owner_rvs - set(names) if exclude and names
            else set(names) if names
            else owner_rvs
        )

        for name in names_to_remove:
            cmd = f"unset {owner}__{name}"
            Database.execute(db=db, cmd=cmd)

        # Remove updated set of RVs or delete owner entry entirely if now empty
        remaining = owner_rvs - names_to_remove
        if remaining:
            Database.rv_names[db][owner] = remaining
        else:
            Database.rv_names[db].pop(owner, None)
            if not Database.rv_names[db]:
                Database.rv_names.pop(db, None)

    @classmethod
    def summarize(cls, db: str, per_attrs: Tuple[str, ...], summaries: Tuple[SumExpr], relation: str = _relation,
                  svar_name: Optional[str] = None) -> RelationValue:
        """
        Full implementation of summarize/summarizeby

        :param db: DB session name
        :param per_attrs:
        :param summaries:
        :param relation:
        :param svar_name:
        :return:
        """
        sum_expr_strings = [f"{s.attr.name} {s.attr.type} {{[{s.expr}]}}" for s in summaries]
        summaries_clause = ' '.join(sum_expr_strings)
        # cmd = (f"set {_relation} [relation summarizeby ${{{relation}}} {{{' '.join(per_attrs)}}} s "
        #        f"{summaries_clause}"
        cmd = (f"set {_relation} [relation summarizeby ${{{relation}}} {{{' '.join(per_attrs)}}} s "
               f"{summaries_clause}]")
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def _expand_expr(cls, cmd_strings: List[str]) -> str:
        """
        Collapse a list of TclRAL command strings into a single string with nested commands.

        For example, these three input cmd_strings:

            0 'relation is ${^relation} subsetof $xactions'
            1 'relation project ${^relation} From_action'
            2 'relation join ${s} $required_inputs'

        Collapse into this returned command string:

            'relation is [relation project [relation join ${s} $required_inputs] From_action] subsetof $xactions'

        Expression expansion is indicated by the _relation variable which contains the TclRAL variable name
        representing the output yield by the previously executed command. As of this writing, the variable name
        is: ${^relation}

        The first command in the list contains the text _relation value somewhere in the string

        We need to replace this value with the next command string surrounded in brackets [<command_string>].
        But, unless that command is the last one in the list, it too will contain the _relation text,
        requiring further expansion.

        So we descend recursively until we hit the bottom command. There we simply return the command text surrounded
        in brackets replacing as we go for each _relation occurence moving back up the stack.

        :param cmd_strings: A list of TclRAL command strings in reverse nesting order
        :return: The fully expanded TclRAL command with all _relation occcurences replaced
        """
        # Ensure that we have either started with at least one command or that we haven't
        # somehow recursed beyond the end of the original command list
        if len(cmd_strings) < 1:
            raise ValueError("At least one command must be specified for expression expansion.")

        # If we have reached the last command, there is no further recursion and we just return the command string
        if len(cmd_strings) == 1:  # Last one, cannot flatten any further, so just return it
            if _relation in cmd_strings[0]:
                raise ValueError(f"Final command: [{cmd_strings[0]}] must not contain substitution marker: {_relation}")
            return cmd_strings[0]

        # There should be exactly one appearance of the _relation
        if _relation in cmd_strings[0]:
            r_expr = cls._expand_expr(cmd_strings[1:])
            expansion = cmd_strings[0].replace(f"${{{_relation}}}", f"[{r_expr}]", 1)
            return expansion
        else:
            raise ValueError(f"Non-final command: [{cmd_strings[0]}] must contain substitution marker: {_relation}")


    @classmethod
    def build_expr(cls, commands) -> str:
        """
        Builds a nested TclRAL expression command for use with summarize

        For example, let's say we want to combine this list of PyRAL Relation commands
        into a single TclRAL expression:

            Relation.join(db=fdb, rname2="required_inputs", rname1="s")
            Relation.project(db=fdb, attributes=("From_action",))
            Relation.set_compare(db=fdb, rname2="xactions", op=SetOp.subset)

        We will receive these as a list of corresponding namedtuples:

            JoinCmd(rname1="s", rname2="required_inputs", attrs=None),
            ProjectCmd(attributes=("From_action",), relation=None),
            SetCompareCmd(rname2="xactions", op=SetOp.subset, rname1=None)


        :param commands:
        :return:
        """

        # Here is each named tuple defining a Relation method that can be built into an expression
        relation_method = {
            "SetCompareCmd": lambda c: cls._cmd_set_compare(rname1=c.rname1, rname2=c.rname2, op=c.op),
            "ProjectCmd": lambda c: cls._cmd_project(relation=c.relation, attributes=c.attributes),
            "JoinCmd": lambda c: cls._cmd_join(rname1=c.rname1, rname2=c.rname2, attrs=c.attrs),
        }

        # Now we create a list of command strings in reverse order
        # This order makes it easy to nest the commands into a single TclRAL command string
        cmd_strings = []
        for c in reversed(commands):
            tuple_name = type(c).__name__
            try:
                cmd = relation_method[tuple_name](c)
            except KeyError:
                raise ValueError(f"Unknown relation command namedtuple: {tuple_name}")
            cmd_strings.append(cmd)

        # Now call this recursive function to perform command substitution and yield our single TclRAL command string
        return cls._expand_expr(cmd_strings)  # TODO: need to specify the db

    @classmethod
    def _cmd_set_compare(cls, rname2: str, op: SetOp, rname1: Optional[str] = None) -> str:
        if rname1 is None:
            rname1 = _relation
        return f'relation is ${{{snake(rname1)}}} {op.value} ${snake(rname2)}'

    @classmethod
    def _cmd_project(cls, attributes, relation: Optional[str] = None) -> str:
        if relation is None:
            relation = _relation
        return f"relation project ${{{snake(relation)}}} {' '.join(attributes)}"

    @classmethod
    def _cmd_union(cls, relations) -> str:
        rvars = [f"${snake(r)}" for r in relations]
        return f'relation union {" ".join(rvars)}'

    @classmethod
    def _cmd_join(cls, rname2: str, attrs, rname1: Optional[str] = None) -> str:
        if rname1 is None:
            rname1 = _relation
        using = f" -using {cls.make_attr_list(attrs)}" if attrs else ""
        return f"relation join ${{{snake(rname1)}}} ${snake(rname2)}{using}"

    @classmethod
    def _cmd_semijoin(cls, rname2: str, attrs, rname1: Optional[str] = None) -> str:
        if rname1 is None:
            rname1 = _relation
        using = f" -using {cls.make_attr_list(attrs)}" if attrs else ""
        return f"relation semijoin ${{{snake(rname1)}}} ${snake(rname2)}{using}"

    @classmethod
    def set_compare(cls, db: str, rname2: str, op: SetOp, rname1: str = _relation) -> bool:
        """

        :param db: DB session name
        :param rname1: If not specified, the previous relation result
        :param rname2: Each rname must have the same header
        :param op: A SetOp enumeration element defined in rtypes.py
        :return: The boolean result of the set operation
        """
        cmd = cls._cmd_set_compare(rname1=snake(rname1), rname2=snake(rname2), op=op)
        result = bool(int(Database.execute(db=db, cmd=cmd)))
        return result

    @classmethod
    def create(cls, db: str, attrs: List[Attribute], tuples: List[namedtuple],
               svar_name: Optional[str] = None) -> RelationValue:
        """
        Create a relation

        :param db: DB session name
        :param attrs: A tuple of attributes (name, type) pairs
        :param tuples: A list of tuples named such that the attributes exactly match the relvar header
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return: Resulting relation as a PyRAL relation value
        """
        h = header(attrs)
        b = body(tuples)
        cmd = f'set {_relation} [relation create {h} {b}]'
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def build_select_expr(cls, selection: str) -> str:
        """
        Convert a Scrall style select expression to an equivalent Tcl string match expression

        For now we only support an and'ed list of direct string matches in the format:

           attr1:str1; attr2:str2, ...

        With the assumption that we would like to select each tuple where

        attr1 == str1 AND attr2 == str2 ...

        We'll convert this to a Tcl expression like this:

        {[string match str1 $attr1] && [string match str2 $attr2] ...}

        Note that this only works for the TclRAL relation restrictwith command and not the
        relation restrict command. But that should suffice for our purposes

        Once our Scrall parser is ready, we can expand the functionality further

        :param selection:  The Scrall style select expression
        :return: The Tcl expression
        """
        # Parse out matches on comma delimiter as a list of strings
        match_strings = selection.split(';')
        # Break each match on the ':' into attr and value as a dictionary
        attr_vals = {a[0].strip(): a[1].strip() for a in [m.split(':') for m in match_strings]}
        # Now build the selection expression from each dictionary item
        sexpr = "{"  # Selection expression is surrounded by brackets
        for attribute, value in attr_vals.items():
            # We AND them all together with the && tcl operator
            sexpr += f"[string match {{{value}}} ${attribute}] && "
        # Remove the trailing && and return the complete selection expression
        return sexpr.rstrip(' &&') + "}"

    @classmethod
    def set_var(cls, db: str, name: str):
        """
        Set a temporary TclRAL relation variable to the most recent returned result.
        This allows us to save a particular TclRAL return value string so that we can plug it
        into a subsequent TclRAL operation.

        :param db: DB session name
        :param name: The variable name (must be a legal Tcl variable name
        """
        session_variable_names.add(name)
        Database.sessions[db].eval(f"set {name} ${{{_relation}}}")

    @classmethod
    def make_attr_list(cls, attrs: Dict[str, str]) -> str:
        """
        Makes a TclRAL attrList to be inserted in a command
        :param attrs:
        :return:
        """
        attr_list = "{"
        for k, v in attrs.items():
            attr_list += f"{snake(k)} {v} "
        return attr_list[:-1] + "}"

    @classmethod
    def semijoin(cls, db: str, rname2: str, attrs: Optional[Dict[str, str]] = None, rname1: str = _relation,
             svar_name: Optional[str] = None) -> RelationValue:
        """
        Perform a semi join on two relations using an optional attribute mapping. If no attributes are specified,
        the semi-join is performed on same named attributes.

        :param db: DB session name
        :param rname1: Name of one relvar to join
        :param rname2: Name of the other relvar
        :param attrs: Dictionary in format { r1.attr_name: r2.attr_name, ... }

        :param db: DB session name
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return Resulting relation as a TclRAL string


        ---
        Copied from the TclRAL man page:

        The semijoin subcommand computes the join of relationValue1 and relationValue2 but eliminates all of the
        attributes of relationValue1 (or alternatively speaking, projecting all attributes of relationValue2).

        The returned relation has a heading the same as relationValue2 and a body consisting of those tuples in
        relationValue2 that would have been included in the natural join with relationValue1.
        As with join, if the -using argument are missing, the join is computed across the attributes in
        relationValue1 and relationValue2 that are named the same. Otherwise the attrList argument is treated
        the same as for the join subcommand.

        Also like the join subcommand, additional relationValue arguments may be given and the result is computed
        in left to right order.

        This implies that the type of the result is always the type of the right most relationValue.
        N.B. the sense of this command is inverted from previous versions of this library.
        """
        if attrs is None:
            attrs = {}
        cmd = f"set {{{_relation}}} [{cls._cmd_semijoin(rname1=rname1, rname2=rname2, attrs=attrs)}]"
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def join(cls, db: str, rname2: str, attrs: Optional[Dict[str, str]] = None, rname1: str = _relation,
             svar_name: Optional[str] = None) -> RelationValue:
        """
        Perform a natural join on two relations using an optional attribute mapping. If no attributes are specified,
        the join is performed on same named attributes.

        :param db: DB session name
        :param rname1: Name of one relvar to join
        :param rname2: Name of the other relvar
        :param attrs: Dictionary in format { r1.attr_name: r2.attr_name, ... }
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return Resulting relation as a TclRAL string
        """
        if attrs is None:
            attrs = {}
        cmd = f"set {{{_relation}}} [{cls._cmd_join(rname1=rname1, rname2=rname2, attrs=attrs)}]"
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def rename(cls, db: str, names: Dict[str, str], relation: str = _relation,
               svar_name: Optional[str] = None) -> RelationValue:
        """
        (NOTE: I only just NOW realized that the TclRAL join command provides an option to specify multiple renames
        as part of a join, because, of course it does! Argghh. So there may not be any need to perform multiple renames
        at once, but hey, there's no harm in providing potentially superfluous functionality as it does at least
        handle the single rename case. - LS)

        Given an input relation, rename one or more attributes from old to new names. This is useful when you want
        to join two relvars on attributes with differing names.

        In SM xUML, it is common for an attribute with one name to reference another attribute of the same
        type but a different name, Employee_ID -> Manager, for exmaple.

        We often need to rename multiple attributes before performing a join, so the single attribute rename
        operation provided by TclRAL is executed once for each element of the names dictionary.

        TclRAL rename syntax:
            relation rename <relationValue> ?oldname newname ...?

        Multiple rename example in TclRAL:
            relation rename ${Attribute_Reference} To_attribute Name
            relation rename ${^relation} To_class Class

        (^relation) is the name of PyRAL's intermediate result session variable, so we are feeding
        the result of the first rename into the next.

        Generated from the PyRAL input:
            relation: 'Attribute_Reference'
            names: {'To_attribute': 'Name', 'To_class': 'Class'}

        :param db: DB session name
        :param relation: The relation to rename
        :param names: Dictionary in format { old_name: new_name }
        :param svar_name:  Name of a TclRAL session variable named for future reference
        :return Resulting relation as a TclRAL string
        """
        r = relation
        result = None
        # Each attribute rename is executed with a separate command
        for old_name, new_name in names.items():
            # The first rename operation is on the supplied relation
            cmd = f'set {_relation} [relation rename ${{{r}}} {old_name} {new_name}]'
            result = Database.execute(db, cmd)
            r = _relation  # Subsequent renames are based on the previous result
        if svar_name:  # Save the final result using the supplied session variable name
            cls.set_var(db, svar_name)
        return cls.make_pyrel(result)  # Result of the final rename (all renames in place)

    @classmethod
    def intersect(cls, db: str, rname2: str, rname1: str = _relation, svar_name: Optional[str] = None
                 ) -> RelationValue:
        """
        Returns the intersection of two relations using the TclRAL intersect command.

        Each relation must be of the same type (same header) as will the result.

        The body of the result consists of those tuples present in both r1 and r2.

        Relational intersection is commutative so the order of the r1 and r2 arguments is not significant.

        The TclRAL syntax is:
            relation intersect <relationValue1> <relationValue2>

        :param db: DB session name
        :param rname1:
        :param rname2:
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return Subtraction relation as a TclRAL string
        """
        cmd = f'set {_relation} [relation intersect ${{{rname1}}} ${rname2}]'
        result = Database.execute(db, cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db, svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def compare(cls, db: str, op: str, rname2: str, rname1: str = _relation) -> bool:
        """
        Returns the intersection of two relations using the TclRAL intersect command.

        Each relation must be of the same type (same header) as will the result.

        The body of the result consists of those tuples present in both r1 and r2.

        Relational intersection is commutative so the order of the r1 and r2 arguments is not significant.

        The TclRAL syntax is:
            relation intersect <relationValue1> <relationValue2>

        :param db: DB session name
        :param op:  Comparison operation
        :param rname1:
        :param rname2:
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return Subtraction relation as a TclRAL string
        """
        cmd = f'relation is ${{{snake(rname1)}}} {op} ${snake(rname2)}'
        result = bool(int(Database.execute(db=db, cmd=cmd)))
        return result


    @classmethod
    def cardinality(cls, db:str, rname: str = _relation) -> int:
        """
        The cardinality subcommand returns the number tuples contained in the body of the relation.

        :param db: DB session name
        :param rname: The tuples in the body of this relation are counted
        :return:
        """
        cmd = f'relation cardinality ${{{snake(rname)}}}'
        result = int(Database.execute(db=db, cmd=cmd))
        return result

    @classmethod
    def subtract(cls, db: str, rname2: str, rname1: str = _relation, svar_name: Optional[str] = None
                 ) -> RelationValue:
        """
        Returns the set difference between two relations using the TclRAL minus command.

        Each relation must be of the same type (same header) as will the result.

        The body of the result consists of those tuples present in r1 but not present in r2.

        Relational subtraction is not commutative so the order of the r1 and r2 arguments is significant.

        The TclRAL syntax is:
            relation minus <relationValue1> <relationValue2>

        TclRAL example taken from the lineage.py Derive method where a set of all classes playing one or more
        subclass roles subtracts all classes playing superclass roles to obtain a set of leaf classes that
        participate as subclasses only.

            relation minus $subs $supers

        Generated from the following PyRAL input:
            rname1: subs
            rname2: supers

        :param tclral: The TclRAL session
        :param rname1: Subtracts value in rname2
        :param rname2: Is subtracted from value in rname1
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return Subtraction relation as a TclRAL string
        """
        cmd = f'set {_relation} [relation minus ${{{rname1}}} ${rname2}]'
        result = Database.execute(db, cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db, svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def get_rval_string(cls, db: str, variable_name: str) -> str:
        """
        Obtain a relation from a TclRAL variable

        :param db: DB session name
        :param variable_name: Name of a variable containing a relation defined in that session
        :return: TclRAL string representing the relation value
        """
        return Database.execute(db, cmd=f"set {variable_name}")

    @classmethod
    def make_pyrel(cls, relation: str, name: str = _relation) -> RelationValue:
        """
        Take a relation obtained from TclRAL and convert it into a pythonic relation value.
        A RelationValue is a named tuple with a header and a body component.
        The header component will be a dictionary of attribute name keys and type values
        The body component will be a list of relational tuples each defined as a dictionary
        with a key matching some attribute of the header and a value for that attribute.

        :param relation: A TclRAL string representing a relation
        :param name: An optional relvar name
        :return: A RelationValue constructed from the provided relation string
        """
        # First check for the dee/dum edge cases
        if relation.strip() == '{} {}':
            # Tabledum (DUM), no attributes and no tuples (relational false value)
            # Header is an empty dictionary and body is an empty list
            return RelationValue(name=name, header={}, body=[])
        if relation.strip() == '{} {{}}':
            # Tabledum (DEE), no attributes and one empty tuple (relational true value)
            # Header is an empty dictionary and body is a list with one empty dictionary element
            return RelationValue(name=name, header={}, body=[{}])

        # Going forward we can assume that there is at least one attribute and zero or more tuples
        h, b = relation.split('}', 1)  # Split at the first closing bracket to obtain header and body strings

        # Construct the header dictionary
        h_items = h.strip('{').split()  # Remove the open brace and split on spaces (no spaces in TclRAL attr names)
        header = dict(zip(h_items[::2], h_items[1::2]))  # Attribute names for keys and TclRAL types for values

        # Construct the body list
        # Each tuple is surrounded by brackets so our first stop is to split them all out into distinct tuple strings
        # Remove leading space and enclosing brackets from entire body
        # 2/-2 skips over leading space and first bracket and truncates final bracket surrounding body
        body = b[2:-2].split('} {')
        body[0] = body[0].lstrip('{')  # Remove preceding bracket from the first tuple

        # Each tuple alternates with the attribute name and the attribute value
        # We want to extract just the values to create the table rows
        # To complicate matters, values may contain spaces. TclRAL attribute names do not.
        # A multi-word value is surrounded by brackets
        # So you might see a tuple like this: Floor_height 32.6 Name {Lower lobby}
        # We need a regex component that will extract the bracketed space delimited values
        # As well as the non-bracketed single word values
        # value_pattern = r"([{}<>\w ]*)"  # Grab a string of any combination of brackets, word characters and spaces
        value_pattern = r"(.*)"  # Grab the whole value string. We'll strip the brackets out later.
        # Now we build this component into an alternating pattern of attribute and value items
        # for the attributes in our relation header
        tuple_pattern = ""
        for a in header.keys():
            tuple_pattern += f"{a} {value_pattern} "
        tuple_pattern = tuple_pattern.rstrip(' ')  # Removes the final trailing space
        # Now we can use the constructed tuple pattern regex to extract a list of values
        # from each row to match our attribute list order
        # Here we apply the tuple_pattern regex to each body row stripping the brackets from each value
        # and end up with a list of unbracketed body row values

        # For tabulate we need a list for the columns and a list of lists for the rows

        # Handle case where there are zero body tuples
        at_least_one_tuple = b.strip('{} ')  # Empty string if no tuples in body
        if not at_least_one_tuple:
            return RelationValue(name=name, header=header, body={})

        # There is at least one body tuple
        if len(header) > 1:
            # More than one column and the regex match returns a convenient tuple in the zero element
            # b_rows = [for row in body]
            b_rows = [[f.strip('{}') for f in re.findall(tuple_pattern, row)[0]] for row in body]
        else:
            # If there is only one match (value), regex returns a string rather than a tuple
            # in the zero element. We need to embed this string in a list
            b_rows = [[re.findall(tuple_pattern, row)[0].strip('{}')] for row in body]
        # Either way, b_rows is a list of lists

        body = [dict(zip(header.keys(), r)) for r in b_rows]
        rval = RelationValue(name=name, header=header, body=body)
        return rval

    @classmethod
    def print(cls, db: str, variable_name: str = _relation, table_name: Optional[str] = None, printout: bool = True) -> str:
        """
        Given the name of a TclRAL relation variable, obtain its value and print it as a table.

        :param db: DB session name
        :param variable_name: Name of the TclRAL variable to print, also used to name the table if no table_name
        :param table_name:  If supplied, this name is used instead of the variable name to name the printed table
        :param printout: Print to console if true
        """
        # convert the TclRAL string value held in the session variable into a PyRAL relation and print it
        rval = cls.make_pyrel(relation=cls.get_rval_string(db=db, variable_name=snake(variable_name)),
                              name=table_name if table_name else variable_name)
        return cls.relformat(rval, printout)


    @classmethod
    def relformat(cls, rval: RelationValue, printout: bool = True) -> str:
        """
        Formats the PyRAL relation into a table and prints it using the imported tabulation module

        :param rval: A PyRAL relation value
        :param printout: Print to console if true
        """
        # Now we have what we need to generate a table
        # Print the relvar name if supplied, otherwise use the default name for the latest result
        tablename = rval.name if rval.name else '<unnamed>'
        if printout:
            print(f"\n-- {tablename} --")
        attr_names = list(rval.header.keys())
        brows = [list(row.values()) for row in rval.body]
        table_text = tabulate(tabular_data=brows, headers=attr_names, tablefmt="outline")
        # That last parameter chooses our table style
        if printout:
            print(table_text)
        return table_text

    @classmethod
    def union(cls, db: str, relations: Tuple[str, ...], svar_name: Optional[str] = None) -> RelationValue:
        """
        The union subcommand returns the set union of two or more relations.
        All relations must be of the same type.

        The result relation has a heading that is the same as any of the arguments and
        has a body consist of all tuples present in any of the relationValue arguments.

        Since the union operation is both associative and commutative, the order of the
        relationValue arguments has no effect the result.

        :param db: DB session name
        :param relations: A tuple providing a
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return Resulting relation as a PyRAL relation value
        """
        relations_s = (snake(t) for t in relations)
        cmd = f'set {_relation} [{cls._cmd_union(relations=relations_s)}]'
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def divide(cls, db: str, dividend: str, divisor: str, mediator: str,
               svar_name: Optional[str] = None) -> RelationValue:
        """
        The divide subcommand implements the relational divide operation.

        The headings of dividend and divisor must be disjoint and the heading of mediator must be
        the union of the dividend and divisor headings.

        The returned result is a new relation that has the same heading as dividend and contains
        all the tuples from dividend whose corresponding tuples in mediator include all the tuples in divisor.
        Stated another way, the result of divide subcommand is the maximal set of tuples from dividend whose
        Cartesian product with divisor is completely contained in mediator.

        :param db: DB session name
        :param dividend: The dividend relation name
        :param divisor: The divisor relation name
        :param mediator: The mediator relation name
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return Resulting relation as a PyRAL relation value
        """
        cmd = f'set {_relation} [relation divide ${dividend} ${divisor} ${mediator}]'
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def raw(cls, db: str, cmd_str: str, relation: str = _relation,
                svar_name: Optional[str] = None) -> RelationValue:
        """
        Passes tcl cmd txt straight through, but uses the variable and relation
        naming mechanism to pipeline input and output like all other commands
        """
        cmd = f'set {_relation} [{cmd_str}]'
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def heading(cls, db:str, relation: str = _relation):
        cmd = f"relation heading ${{{relation}}}"
        result = Database.execute(db=db, cmd=cmd)
        return result

    @classmethod
    def project(cls, db: str, attributes: Tuple[str, ...], exclude: bool = False, relation: str = _relation,
                svar_name: Optional[str] = None) -> RelationValue:
        """
        Returns a relation whose heading consists of only a set of selected attributes.
        The body of the result consists of the corresponding tuples from the specified relation,
        removing any duplicates created by considering only a subset of the attributes.

        :param db: DB session name
        :param attributes: Attributes to be projected
        :param relation: The relation to be projected
        :param exclude: If true, all attributes will be returned except for those in the attributes tuple
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return Resulting relation as a PyRAL relation value
        """
        # Create a list of attributes to project by inclusion or exclusion
        if exclude:
            # Use heading method to get all attributes defined on the relation
            # and then project on all of these except those attributes provided in the tuple
            attr_types = Relation.heading(db=db, relation=relation)
            # heading returns a string delimited by spaces with attribute type pairs
            tokens = attr_types.split()
            # Now skip over all the type names and just grab the attribute names (1st, 3rd, ...)
            pairs = zip(tokens[::2], tokens[1::2])
            # Exclude the provided attribute names
            project_attrs = [name for name, _ in pairs if name not in attributes]
        else:
            # Just use the provided attributes
            project_attrs = list(attributes)

        attributes_s = (snake(s) for s in project_attrs)
        cmd = f"set {_relation} [{cls._cmd_project(relation=snake(relation), attributes=attributes_s)}]"
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def make_comparison(cls, attr_name: str, values: set | str) -> str:
        if isinstance(values, set):
            vmatch = [f"[string match {{{v}}} [tuple extract $t {attr_name}]]" for v in values]
            return '(' + ' || '.join(vmatch) + ')'

        # There's only one value
        return f"[string match {{{values}}} [tuple extract $t {attr_name}]]"

    @classmethod
    def summarizeby(cls, db: str, relation: str, attrs: List[str], sum_attr: Attribute, op='count',
                    svar_name: Optional[str] = None) -> RelationValue:
        """
        DEPRECATED

        Only one summarization operation supported by PyRAL at the moment - count
        :param db: DB session name
        :param relation:
        :param attrs:
        :param sum_attr:
        :param op:
        :param svar_name:
        :return:

        From TclRAL man page:
        --
        The summarizeby subcommand is a more convenient form of summarize where the per relation is a projection of
        the relation value that is to be summarized. Rather than supplying a per relation, instead a list of
        attributes is given by the attrList argument and relationValue is projected on those attributes and
        used as the per relation. The arguments and results are otherwise the same as for the summarize command.

        TclRAL syntax:
        relation summarizeby relationValue attrList relationVarName attr type expression ?attr type expression ...?

        TclRAL command example:
        % relformat [relation summarizeby $OWNERSHIP Acquired s NumAcquired int {[relation cardinality $s]}]

        PyRal implements this in limited form for now.  Only the cardinality operations is supported.
        Given this input to PyRAL:

            Relation.summarizeby(db=stdb, relation='Region', attrs=['Data_box', 'Title_block_pattern'],
                                 sum_attr=Attribute(name='Qty', type='int'), svar_name="Number_of_regions")

            We get the TclRAL command:
                relation summarizeby ${Region} {Data_box Title_block_pattern} s Qty int {[relation cardinality $s]}

        Applied to this example relation:

        -- Region --
        +------------+-----------------------+---------------+
        |   Data_box | Title_block_pattern   |   Stack_order |
        +============+=======================+===============+
        |          3 | Complex               |             1 |
        |          3 | Complex               |             2 |
        |          3 | SE Simple             |             1 |
        |          5 | SE Simple             |             1 |
        |          6 | SE Simple             |             1 |
        |          6 | SE Simple             |             2 |
        |          7 | SE Simple             |             1 |
        |          7 | SE Simple             |             2 |
        +------------+-----------------------+---------------+

        We get the following result which tells us the number of Regions per Data Box+Title_block_pattern

        -- Number_of_regions --
        +------------+-----------------------+-------+
        |   Data_box | Title_block_pattern   |   Qty |
        +============+=======================+=======+
        |          3 | Complex               |     2 |
        |          3 | SE Simple             |     1 |
        |          5 | SE Simple             |     1 |
        |          6 | SE Simple             |     2 |
        |          7 | SE Simple             |     2 |
        +------------+-----------------------+-------+

        --
        """
        cmd = (f"set {_relation} [relation summarizeby ${{{relation}}} {{{' '.join(attrs)}}} s "
               f"{sum_attr.name} {sum_attr.type} {{[relation cardinality $s]}}]")
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:  # Save the result using the supplied session variable name
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def rank(cls, db: str, order: Order, sort_attr_name: str, rank_attr_name: str = _RANK, relation: str = _relation,
             svar_name: Optional[str] = None) -> RelationValue:
        """
        TclRAL documentation and syntax:

            relation rank relationValue ?-ascending | -descending? rankAttr newAttr

        The rank subcommand returns a new relation whose heading is the same as relationValue
        extending by an attribute named newAttr. The type of newAttr will be int and its value
        will be set to the number of tuples in relationValue where the value of rankAttr is
        less than or equal to (?-descending?) or greater than or equal to (?-ascending?) that of
        the given tuple. The default ranking is -ascending.

        The type of rankAttr must be int, double, or string.

        The rank command is useful when it is desirable to limit the number of tuples in the result.

        PyRAL example:

            result = Relation.rank(db=ev, relation="shafts_rv", sort_attr_name="Speed", order=Order.DESCENDING)

        :param db: DB session name
        :param order: ascending or descending (no default ordering)
        :param sort_attr_name: The values of this attribute will be sorted (TclRAL rankAttr)
        :param rank_attr_name: The name of the added rank number attribute (TclRAL newAttr)
        :param relation: Relation to be sorted
        :param svar_name: An optional session variable that holds the result
        :return: Relation with added rank_attr_name
        """
        cmd = f"set {_relation} [relation rank ${{{relation}}} -{order.value}{{{sort_attr_name}}}{{{rank_attr_name}}}]"

        result = Database.execute(db=db, cmd=cmd)
        if svar_name:
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def tag(cls, db: str, tag_attr_name: str = _TAG, sort_attrs: Tuple[str, ...] = None,
            order: Order = Order.ASCENDING, relation: str = _relation,
            svar_name: Optional[str] = None) -> RelationValue:
        """
        TclRAL documentation:

            relation tag relationValue attrName ?-ascending | -descending sort-attr-list? ?-within attr-list?

        The tag subcommand creates a new relation which has the same heading as relationValue extended by a
        new attribute named attrName. The type of attrName will be int and will have the values between 0 and the
        cardinality of relationValue minus one. The tuples in relationValue will be extended in either ascending or
        descending order of the sort-attr-list.

        If no sort-attr-list argument is given, then the tagging order is arbitrary.

        If the ?-within? argument is given then the values of attrName attribute will be unique within the subset
        of tuples which match the values of attr-list. The tag command is useful when a full ordering needs to be
        placed on a relation. For example, tagging a relation will allow projecting both the tag and another
        attribute without losing any values.

        PyRAL implements this partially for now and the -within option is not yet supported.

        :param sort_attrs:
        :param order:
        :param db:
        :param tag_attr_name:
        :param relation:
        :param svar_name:
        :return:
        """
        do_sort = "" if not sort_attrs else f" -{order.value} {' '.join(sort_attrs)} "
        cmd = f"set {_relation} [relation tag ${{{snake(relation)}}} {{{tag_attr_name}}} {do_sort}]"
        result = Database.execute(db=db, cmd=cmd)
        if svar_name:
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)

    @classmethod
    def rank_restrict(cls, db: str, attr_name: str, extent: Extent, card: Card, relation: str = _relation,
                      svar_name: Optional[str] = None) -> RelationValue:
        """
        Our goal here is to order the tuples of a relation on some attribute and then select the
        furthest extent, either greatest or least.

        For example, we might want to select the highest flying aircraft.

        We do this by ranking all aircraft tuples according to altitude in descending order.

        Then we select all those of ranking 1. Since multiple aircraft might be flying at the same highest
        altitude, there may be multiple ranked as 1. And if there are no aircraft instances, we'll find
        no highest flying aicraft.

        If the user specifies the ALL cardinality, they could obtain multiple aircraft tuples at the same
        altitude. However, the ONE cardinality specifies that only one will be selected and the user cannot
        choose which particular tuple is selected.

        :param db: DB session name
        :param attr_name: Tuples are ordered based on values of this attribute
        :param extent: greatest or least
        :param card: one or all
        :param relation: Selection is on tuples of this relation
        :param svar_name: Relation result is stored in this optional TclRAL variable for subsequent operations to use
        :return: The tuple or tuples at the same extent
        """
        order = Order.DESCENDING if extent == Extent.GREATEST else Order.ASCENDING
        if card == Card.ALL:
            Relation.rank(db=db, order=order, sort_attr_name=attr_name, relation=relation)
            R = f"{_RANK}:1"
            Relation.restrict(db=db, restriction=R)
            return Relation.project(db=db, attributes=(_RANK,), exclude=True, svar_name=svar_name)
        else:  # Card must be ONE
            Relation.tag(db=db, order=order, sort_attrs=(attr_name,), relation=relation)
            R = f"{_TAG}:0"
            Relation.restrict(db=db, restriction=R)
            return Relation.project(db=db, attributes=(_TAG,), exclude=True, svar_name=svar_name)

    @classmethod
    def restrict(cls, db: str, restriction: Optional[str] = None, relation: str = _relation,
                  svar_name: Optional[str] = None) -> RelationValue:
        """
        Here we select zero or more tuples that match the supplied criteria.

        In relational theory this is known as a restriction operation.

        TclRAL syntax:
            relation restrictwith <relationValue> <expression>

        The most common usage scenario is to select on a single identifier attribute value.

            R = f"ID:S1"
            result = Relation.restrict(db=ev, relation="shafts_rv", restriction=R, svar_name="restriction")

            -- restriction --
            +------+---------+--------------+
            | ID   |   Speed | In_service   |
            +======+=========+==============+
            | S1   |      31 | True         |
            +------+---------+--------------+

            This yields the TclRAL statement

            relation restrictwith ${shafts_rv} {[string match {S1} $ID]}

            So we have the command, a relation variable, and the string match.
            The string match arguments are the value to match and the name of the attribute as a tcl variable

        If there is any whitespace in the value, use <> brackets to enclose it. For example:

            R = f"Level_name:<3rd Floor>"

        Note that the attribute names must be specified in snake case. Relation names can have spaces, but are
        automatically converted to snake case in the generated TclRAL. It's a bit trickier to do this with all the
        attribute names since they are embedded in the restriction expression, so we put that burden on
        the user.

        Note that the : symbol is shorthand for matching and only works with strings. The == operator can
        only be used to match numeric values.

        You can match multiple values using the ', ' to AND them together. Here is the supplied PyRAL expression
        and generated TclRAL expression:

            PyRAL: R = f"ID:<{i}>, In_service:<{True}>"
            TclRAL: {[string match {S1} $ID] && [string match {True} $In_service]}

        Here we use i to specify the "S1" ID value using the python formatted string {} brackets.
        The brackets around True are necessary to ensure a boolean value is inserted and not the "True" string.
        (though the string value would work as well since Tcl represents a boolean value as a truthy string)

        Numeric comparisons can be supplied using the usual operators with s set to the integer 14:

            PyRAL: R = f"Speed > {s}"
            TclRAL: {[expr {$Speed > 14}]}

        In a numeric comparison you do not need to surround the numeric value with <> brackets.

        Finally, you can express more complex logic using nested parentheses and the logic operators
        AND, OR, NOT as follows:

            PyRAL: R = f"ID:<{v}> OR (In_service:<{True}> AND Speed > {s})"
            TclRAL: {[string match {S1} $ID] || ([string match {True} $In_service] && [expr {$Speed > 31}])}

        :param db: DB session name
        :param relation: Name of a relation variable where the operation is applied
        :param restriction: A string in Scrall notation that specifies the restriction criteria
        :param svar_name: An optional session variable that holds the result
        :return: The TclRAL string result representing the restricted tuple set
        """
        relation_s = snake(relation)
        if not restriction:
            cmd = f"set {_relation} [set {relation_s}]"
        else:
            # Handle arithmetic comparisons like Speed > 14
            restrict_tcl = re.sub(
                pattern=r'(\w+)\s*(==|!=|>=|<=|<|>)\s*(-?\d+(?:\.\d+)?)',
                repl=r'[expr [tuple extract $t \1] \2 \3]',
                string=restriction
            )

            # Handle attr:<value> form for string match
            restrict_tcl = re.sub(
                pattern=r'([\w_]+):<([^>]*)>',
                repl=r'[string match {\2} [tuple extract $t \1]]',
                string=restrict_tcl
            )

            # Handle attr:value form for string match (NEW addition)
            restrict_tcl = re.sub(
                pattern=r'([\w_]+):([^\s<>()&|!]+)',
                repl=r'[string match {\2} [tuple extract $t \1]]',
                string=restrict_tcl
            )

            # Convert boolean logic operators and NOT
            restrict_tcl = restrict_tcl.replace(' OR ', ' || ') \
                .replace(', ', ' && ') \
                .replace(' AND ', ' && ') \
                .replace('NOT ', '!')

            rexpr = f"{{{restrict_tcl}}}"
            cmd = f"set {_relation} [relation restrict ${{{relation_s}}} t {rexpr}]"

        result = Database.execute(db=db, cmd=cmd)
        if svar_name:
            cls.set_var(db=db, name=svar_name)
        return cls.make_pyrel(result)




""" test_tr.py - test transactions """

import pytest
from pyral.relation import Relation
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.transaction import Transaction
from pyral.rtypes import Attribute, RelationValue

from collections import namedtuple

Aircraft_i = namedtuple('Aircraft_i', 'ID Altitude Heading')
Pilot_i = namedtuple('Pilot_i', 'Callsign Tailnumber Age')

@pytest.fixture(autouse=True)
def before_after_tests(tmpdir):
    acdb = "ac"
    Database.open_session(acdb)

    yield

    Database.close_session(acdb)
    return acdb

def test_tr1():
    acdb = "ac"
    tr1 = Transaction.open(acdb, "tr1")
    Relvar.create_relvar(acdb, name='Aircraft', attrs=[Attribute('ID', 'string'), Attribute('Altitude', 'int'),
                                                       Attribute('Heading', 'int')], ids={1: ['ID']})
    Relvar.insert(acdb, tr=tr1, relvar='Aircraft', tuples=[
        Aircraft_i(ID='N1397Q', Altitude=13275, Heading=320),
        Aircraft_i(ID='N1309Z', Altitude=10100, Heading=273),
        Aircraft_i(ID='N5130B', Altitude=8159, Heading=90),
    ])

    Relvar.create_relvar(acdb, name='Pilot', attrs=[Attribute('Callsign', 'string'), Attribute('Tailnumber', 'string'),
                                                    Attribute('Age', 'int')], ids={1: ['Callsign']})
    Relvar.insert(acdb, tr=tr1, relvar='Pilot', tuples=[
        Pilot_i(Callsign='Viper', Tailnumber='N1397Q', Age=22),
        Pilot_i(Callsign='Joker', Tailnumber='N5130B', Age=31),
    ])
    Transaction.execute(acdb, tr1)
    assert Database.names(acdb) == '::Pilot ::Aircraft'


import pytest
from pyral.relation import Relation
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.rtypes import Attribute, RelationValue
from collections import namedtuple

Fixed_Wing_Aircraft_i = namedtuple('Fixed_Wing_Aircraft_i', 'ID Altitude Compass_heading')
Pilot_i = namedtuple('Pilot_i', 'Callsign Tail_number Age')

@pytest.fixture(scope='module', autouse=True)
def tear_down():
    yield
    Database.close_session("ac")

@pytest.fixture(scope='module')
def aircraft_db():
    acdb = "ac"
    Database.open_session(acdb)

    Relvar.create_relvar(
        db=acdb,
        name='Fixed Wing Aircraft',
        attrs=[
            Attribute('ID', 'string'),
            Attribute('Altitude', 'int'),
            Attribute('Compass heading', 'int')
        ],
        ids={1: ['ID']}
    )
    Relvar.insert(db=acdb, relvar='Fixed Wing Aircraft', tuples=[
        Fixed_Wing_Aircraft_i(ID='N1397Q', Altitude=13275, Compass_heading=320),
        Fixed_Wing_Aircraft_i(ID='N1309Z', Altitude=10100, Compass_heading=273),
        Fixed_Wing_Aircraft_i(ID='N5130B', Altitude=8159, Compass_heading=90),
    ])

    Relvar.create_relvar(
        db=acdb,
        name='Pilot',
        attrs=[
            Attribute('Callsign', 'string'),
            Attribute('Tail_number', 'string'),
            Attribute('Age', 'int')
        ],
        ids={1: ['Callsign']}
    )
    Relvar.insert(db=acdb, relvar='Pilot', tuples=[
        Pilot_i(Callsign='Viper', Tail_number='N1397Q', Age=22),
        Pilot_i(Callsign='Joker', Tail_number='N5130B', Age=31),
    ])

    return acdb

def test_string_match_angle_brackets(aircraft_db):
    result = Relation.restrict(db=aircraft_db, relation='Pilot', restriction="Callsign:<Viper>")
    assert len(result.body) == 1
    assert result.body[0]['Callsign'] == 'Viper'

def test_string_match_colon_value(aircraft_db):
    result = Relation.restrict(db=aircraft_db, relation='Pilot', restriction="Callsign:Viper")
    assert len(result.body) == 1
    assert result.body[0]['Tail_number'] == 'N1397Q'

def test_numeric_comparison(aircraft_db):
    result = Relation.restrict(db=aircraft_db, relation='Fixed Wing Aircraft', restriction="Altitude > 10000")
    ids = {row['ID'] for row in result.body}
    assert ids == {'N1397Q', 'N1309Z'}

def test_logical_and(aircraft_db):
    result = Relation.restrict(
        db=aircraft_db,
        relation='Fixed Wing Aircraft',
        restriction="Altitude > 10000, Compass_heading:320"
    )
    ids = {row['ID'] for row in result.body}
    assert ids == {'N1397Q'}

def test_logical_or(aircraft_db):
    result = Relation.restrict(
        db=aircraft_db,
        relation='Fixed Wing Aircraft',
        restriction="ID:<N1309Z> OR ID:<N5130B>"
    )
    ids = {row['ID'] for row in result.body}
    assert ids == {'N1309Z', 'N5130B'}

def test_parentheses_and_or(aircraft_db):
    restriction = "(ID:<N1397Q>, Altitude > 13000) OR (ID:<N1309Z>, Altitude > 9000)"
    result = Relation.restrict(db=aircraft_db, relation='Fixed Wing Aircraft', restriction=restriction)
    ids = {row['ID'] for row in result.body}
    assert ids == {'N1397Q', 'N1309Z'}

def test_not_logic(aircraft_db):
    result = Relation.restrict(db=aircraft_db, relation='Pilot', restriction="NOT Age:<22>")
    callsigns = {row['Callsign'] for row in result.body}
    assert callsigns == {'Joker'}


"""
print_table.py -- Test table printing

"""
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction
from pyral.rtypes import Attribute

from collections import namedtuple

Aircraft_i = namedtuple('Aircraft_i', 'ID Altitude Heading')
Pilot_i = namedtuple('Pilot_i', 'Callsign Tailnumber Age')
Region_i = namedtuple('Region_i', 'Data_box Title_block_pattern Stack_order')
DataBox_i = namedtuple('DataBox_i', 'ID Pattern Name')

class SumTest:

    """
    Test summarizeby command for count / cardinality
    """
    @classmethod
    def do_r(cls):
        stdb = "stdb"

        Database.open_session(name=stdb)
        Relvar.create_relvar(db=stdb, name='Region', attrs=[
            Attribute('Data_box', 'int'),
            Attribute('Title_block_pattern', 'string'),
            Attribute('Stack_order', 'int')],ids={1: ['Data_box', 'Title_block_pattern', 'Stack_order']})

        Relvar.insert(db=stdb, relvar='Region', tuples=[
            Region_i(Data_box=3, Title_block_pattern='Complex', Stack_order=1),
            Region_i(Data_box=3, Title_block_pattern='Complex', Stack_order=2),
            Region_i(Data_box=3, Title_block_pattern='SE Simple', Stack_order=1),
            Region_i(Data_box=5, Title_block_pattern='SE Simple', Stack_order=1),
            Region_i(Data_box=6, Title_block_pattern='SE Simple', Stack_order=1),
            Region_i(Data_box=6, Title_block_pattern='SE Simple', Stack_order=2),
            Region_i(Data_box=7, Title_block_pattern='SE Simple', Stack_order=1),
            Region_i(Data_box=7, Title_block_pattern='SE Simple', Stack_order=2),
        ])
        Relvar.printall(stdb)

        result = Relation.summarizeby(db=stdb, relation='Region', attrs=['Data_box', 'Title_block_pattern'],
                                      sum_attr=Attribute(name='Qty', type='int'), svar_name="Number_of_regions")
        Relation.print(db=stdb, variable_name="Number_of_regions")
        pass

class TableTest:

    @classmethod
    def do_r(cls):
        acdb = "acdb"

        Database.open_session(name=acdb)
        Relvar.create_relvar(acdb, name='Aircraft', attrs=[Attribute('ID', 'string'), Attribute('Altitude', 'int'),
                                                           Attribute('Heading', 'int')], ids={1: ['ID']})
        Relvar.create_relvar(acdb, name='Pilot',
                             attrs=[Attribute('Callsign', 'string'), Attribute('Tailnumber', 'string'),
                                    Attribute('Age', 'int')], ids={1: ['Callsign']})
        tr_p = Transaction.open(acdb, "Pilot")
        Relvar.insert(acdb, tr=tr_p, relvar='Pilot', tuples=[
            Pilot_i(Callsign='Viper', Tailnumber='N1397Q', Age=22),
            Pilot_i(Callsign='Joker', Tailnumber='N5130B', Age=31),
        ])
        tr_a = Transaction.open(acdb, "Aircraft")
        Relvar.insert(acdb, tr=tr_a, relvar='Aircraft', tuples=[
            Aircraft_i(ID='N1397Q', Altitude=13275, Heading=320),
            Aircraft_i(ID='N1309Z', Altitude=10100, Heading=273),
            Aircraft_i(ID='N5130B', Altitude=8159, Heading=90),
        ])

        Transaction.execute(acdb, tr_p)
        Relation.print(acdb, "Pilot")
        Transaction.execute(acdb, tr_a)
        Relation.print(acdb, "Aircraft")
        result = Relation.restrict(acdb, relation='Aircraft')
        pass

        # aone = Relvar.select_id(acdb, 'Aircraft', {'ID': '1397Q'}, svar_name='One')
        # Relation.relformat(aone)

        # result = Relation.join(acdb, rname1='Pilot', rname2='Aircraft')
        #
        # # result = Relation.project(db, attributes=('Age',), relation='Pilot')
        # Relation.relformat(result)
        #
        # a = Relation.restrict(db, restriction=f"Altitude:<10100>", relation="Aircraft")
        # b = Relation.restrict(db, restriction=f"Altitude:<10100>", relation="Aircraft")
        # Relation.relformat(a)

        # db.eval('set high [relation restrict $Aircraft t {[tuple extract $t Altitude] > 9000}]' )
        # Relation.print(db, 'high')
        # db.eval('set low [relation restrict $Aircraft t {[tuple extract $t Altitude] < 13000}]' )
        # Relation.print(db, 'low')
        # #
        # b = Relation.intersect(db, rname2='high', rname1='low')
        # Relation.relformat(b)
        #
        # thesame = db.eval('relation is $Aircraft != $Aircraft')
        # print(thesame)
        #
        # thesame = Relation.compare(db, op='<=', rname1='Aircraft', rname2='Aircraft')
        # print(thesame)

        # db.eval('set between [relation intersect $high $low]' )
        # Relation.print(db, 'between')

        # lower = Relation.subtract(db, rname2='r', rname1='Aircraft')
        # Relation.relformat(lower)



from collections import namedtuple
from pyral.rtypes import header, body, Attribute

# Test fixtures
Waypoint = namedtuple('Waypoint', ['WPT_number', 'Lat', 'Long', 'Frequency'])

def test_header_single_attribute():
    attrs = [Attribute(name='ID', type='int')]
    assert header(attrs) == "{ID int}"

def test_header_multiple_attributes():
    attrs = [
        Attribute(name='WPT number', type='int'),
        Attribute(name='Lat', type='string'),
        Attribute(name='Long', type='string'),
        Attribute(name='Frequency', type='double')
    ]
    expected = "{WPT_number int Lat string Long string Frequency double}"
    assert header(attrs) == expected

def test_body_single_tuple():
    t = Waypoint(WPT_number=3, Lat="37° 46' 30\" N", Long="-122° 25' 10\"", Frequency="117.95")
    expected = '{WPT_number {3} Lat {37° 46\' 30" N} Long {-122° 25\' 10"} Frequency {117.95}}'
    assert body([t]) == expected

def test_body_multiple_tuples():
    t1 = Waypoint(WPT_number=1, Lat="35° 12'", Long="-118° 22'", Frequency="113.20")
    t2 = Waypoint(WPT_number=2, Lat="36° 10'", Long="-119° 15'", Frequency="115.10")
    expected = (
        '{WPT_number {1} Lat {35° 12\'} Long {-118° 22\'} Frequency {113.20}} '
        '{WPT_number {2} Lat {36° 10\'} Long {-119° 15\'} Frequency {115.10}}'
    )
    assert body([t1, t2]) == expected

def test_body_empty():
    assert body([]) == ""



"""
restrict_play_ac.py -- Test advanced restriction with aircraft example

"""
# System
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute

Fixed_Wing_Aircraft_i = namedtuple('Fixed_Wing_Aircraft_i', 'ID Altitude Compass_heading')
Pilot_i = namedtuple('Pilot_i', 'Callsign Tail_number Age')

acdb = "ac"

def play():
    Database.open_session(acdb)

    Relvar.create_relvar(
        db=acdb,
        name='Fixed Wing Aircraft',
        attrs=[
            Attribute('ID', 'string'),
            Attribute('Altitude', 'int'),
            Attribute('Compass heading', 'int')
        ],
        ids={1: ['ID']}
    )
    Relvar.insert(db=acdb, relvar='Fixed Wing Aircraft', tuples=[
        Fixed_Wing_Aircraft_i(ID='N1397Q', Altitude=13275, Compass_heading=320),
        Fixed_Wing_Aircraft_i(ID='N1309Z', Altitude=10100, Compass_heading=273),
        Fixed_Wing_Aircraft_i(ID='N5130B', Altitude=8159, Compass_heading=90),
    ])

    Relvar.create_relvar(
        db=acdb,
        name='Pilot',
        attrs=[
            Attribute('Callsign', 'string'),
            Attribute('Tail_number', 'string'),
            Attribute('Age', 'int')
        ],
        ids={1: ['Callsign']}
    )
    Relvar.insert(db=acdb, relvar='Pilot', tuples=[
        Pilot_i(Callsign='Viper', Tail_number='N1397Q', Age=22),
        Pilot_i(Callsign='Joker', Tail_number='N5130B', Age=31),
    ])

    # R = f"Altitude > 10000"
    # R = f"Compass_heading:320"
    # R = f"Altitude > 10000, Compass_heading:320"
    R = f"Compass_heading > 90, Compass_heading < 320"
    result = Relation.restrict(db=acdb, relation="Fixed Wing Aircraft", restriction=R, svar_name="restriction")
    Relation.print(db=acdb, variable_name="restriction")

    Database.close_session(acdb)

    pass



"""
PyRAL

"""

# System
import logging
import logging.config
import sys
import argparse
from pathlib import Path

# PyRAL
from pyral import version

_logpath = Path("pyral.log")


def get_logger():
    """Initiate the logger"""
    log_conf_path = Path(__file__).parent / 'log.conf'  # Logging configuration is in this file
    logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
    return logging.getLogger(__name__)  # Create a logger for this module


# Configure the expected parameters and actions for the argparse module
def parse(cl_input):
    parser = argparse.ArgumentParser(description='PyRAL')
    parser.add_argument('-T', '--test', action='store_true',
                        help='Run print test'),
    parser.add_argument('-R', '--rebuild', action='store_true',
                        help='Rebuild the database.'),
    parser.add_argument('-D', '--debug', action='store_true',
                        help='Debug mode'),
    parser.add_argument('-V', '--version', action='store_true',
                        help='Print the current version of PyRAL')
    return parser.parse_args(cl_input)


def main():
    # Start logging
    logger = get_logger()
    logger.info(f'PyRAL version: {version}')

    # Parse the command line args
    args = parse(sys.argv[1:])

    if args.version:
        # Just print the version and quit
        print(f'PyRAL version: {version}')

    # if args.test:
    #     from print_table import SumTest
    #     SumTest.do_r()
    if args.test:
        # from print_table import TableTest
        # from pyral.experiments.sum_play import SumTest
        # from pyral.experiments.sum_play2 import SumTest2
        # from pyral.experiments.union_play import play
        # from pyral.experiments.join_play import play
        from pyral.experiments.restrict_play import play
        # from pyral.experiments.setops_play import SetPlay
        # SetPlay.setup()
        play()
        # from pyral.experiments.dogs_example import Dogs
        # Dogs.setup()

    logger.info("No problemo")  # We didn't die on an exception, basically
    print("\nNo problemo")


if __name__ == "__main__":
    main()



"""
exceptions.py – PyRAL exceptions
"""

class PyRALException(Exception):
    """ Top level PyRAL exception """
    pass

class Transaction(PyRALException):
    """ Top level PyRAL Transaction exception """
    pass


class SessionNotOpen(PyRALException):
    """ The database session has not been opened """
    pass

class TclRALException(PyRALException):
    """ A TclRAL or Tcl error occured inside the tcl interpeter """
    pass

class DuplicateTransaction(Transaction):
    """ Cannot use name to open a new transaction since one is aready open by that name """
    pass

class UnNamedTransaction(Transaction):
    """ Name does not correspond to any pending transaction """
    pass

class NoOpenTransaction(PyRALException):
    """ Attempt to add statement when no transaction has been opened """
    pass

class RestrictOneOnZeroCardinality(PyRALException):
    """ Attempted restrict-one operation on relation with no tuples """
    pass



# test database management

import pytest
from pyral.relation import Relation
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.rtypes import Attribute

from collections import namedtuple

Fixed_Wing_Aircraft_i = namedtuple('Fixed_Wing_Aircraft_i', 'ID Altitude Compass_heading')
Pilot_i = namedtuple('Pilot_i', 'Callsign Tail_number Age')

@pytest.fixture(scope='module', autouse=True)
def tear_down():
    yield
    Database.close_session("ac")

@pytest.fixture(scope='module')
def aircraft_db():
    acdb = "ac"

    Database.open_session(acdb)
    Relvar.create_relvar(db=acdb, name='Fixed Wing Aircraft', attrs=[Attribute('ID', 'string'), Attribute('Altitude', 'int'),
                                                       Attribute('Compass heading', 'int')], ids={1: ['ID']})
    Relvar.insert(db=acdb, relvar='Fixed Wing Aircraft', tuples=[
        Fixed_Wing_Aircraft_i(ID='N1397Q', Altitude=13275, Compass_heading=320),
        Fixed_Wing_Aircraft_i(ID='N1309Z', Altitude=10100, Compass_heading=273),
        Fixed_Wing_Aircraft_i(ID='N5130B', Altitude=8159, Compass_heading=90),
    ])

    Relvar.create_relvar(db=acdb, name='Pilot', attrs=[Attribute('Callsign', 'string'), Attribute('Tail_number', 'string'),
                                                    Attribute('Age', 'int')], ids={1: ['Callsign']})
    Relvar.insert(db=acdb, relvar='Pilot', tuples=[
        Pilot_i(Callsign='Viper', Tail_number='N1397Q', Age=22),
        Pilot_i(Callsign='Joker', Tail_number='N5130B', Age=31),
    ])
    return acdb

def test_rv_declare1(aircraft_db):
    join_example = Relation.declare_rv(db=aircraft_db, owner="P1", name="xjoin")
    Relation.join(db=aircraft_db, rname1="Pilot", rname2="Fixed Wing Aircraft", attrs={"Tail number": "ID"},
                  svar_name=join_example)
    assert Database.rv_names == {'ac': {'P1': {'xjoin'}}}
    Relation.free_rvs(db=aircraft_db, owner="P1")
    assert Database.rv_names == {}

def test_rv_declare2(aircraft_db):
    join_example = Relation.declare_rv(db=aircraft_db, owner="P1", name="xjoin")
    Relation.join(db=aircraft_db, rname1="Pilot", rname2="Fixed Wing Aircraft", attrs={"Tail number": "ID"},
                  svar_name=join_example)
    semijoin_example = Relation.declare_rv(db=aircraft_db, owner="P1", name="xsemijoin")
    Relation.semijoin(db=aircraft_db, rname1="Pilot", rname2="Fixed Wing Aircraft", attrs={"Tail_number": "ID"},
                      svar_name=semijoin_example)
    assert Database.rv_names == {'ac': {'P1': {'xjoin', 'xsemijoin'}}}
    Relation.free_rvs(db=aircraft_db, owner="P1")
    assert Database.rv_names == {}

def test_rv_free_bad_owner(aircraft_db):
    with pytest.raises(KeyError):
        Relation.free_rvs(db=aircraft_db, owner="P2")

def test_rv_bad_session(aircraft_db):
    with pytest.raises(KeyError):
        Relation.declare_rv(db="oink", owner="P1", name="xjoin")

def test_rv_duplicate_name(aircraft_db):
    Relation.declare_rv(db=aircraft_db, owner="P1", name="xjoin")
    with pytest.raises(KeyError):
        Relation.declare_rv(db=aircraft_db, owner="P1", name="xjoin")
